/**
 * Copyright (C) 2007-2011 Linutop SAS <http://www.linutop.com/>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>

#include "main.h"
#include "wk-browser.h"
#include "wk-pl-parser.h"
#include "wk-slideshow-item.h"

#define DEFAULT_ITEM_TIMEOUT 5

static int cur_priority;
static int cur_id;
static int prev_id = -1;
static gboolean in_loop = FALSE;
static gboolean idle = FALSE;

typedef struct {
    int start_year;
    int start_month;
    int start_day;
    char start_date[11];

    int end_year;
    int end_month;
    int end_day;
    char end_date[11];

    int start_hour;
    int start_min;
    char start_time[6];

    int end_hour;
    int end_min;
    char end_time[6];

    int weekdays;
} WkPlLoopInfo;

void wk_pl_info_print (WkPlLoopInfo *info);

static void
append_dir_to_list (const gchar *dirname, gint t, GSList **l)
{
    GDir *dir;
    gchar *filename = NULL, *uri = NULL;
    const gchar *localname = NULL;
    WkSlideshowItem *item;
    WkSlideshowItemType type;

    if ((dir = g_dir_open (dirname, 0, NULL)) == NULL)
        return;

    while ((localname = g_dir_read_name (dir)) != NULL)
    {
        uri = g_strdup_printf ("file://%s/%s", dirname, localname);
        filename = uri+7;

        if (g_file_test (filename, G_FILE_TEST_IS_DIR))
        {
            append_dir_to_list (filename, t, l);
            g_free (uri);
        }
        else
        {
            type = wk_slideshow_item_guess_type (uri);
            item = wk_slideshow_item_new (uri, type, t, FALSE);
            g_free (uri);
            *l = g_slist_append (*l, item);
        }
    }
    g_dir_close (dir);
}

GSList *
old_school_playlist_parser (const gchar *str)
{
    WkSlideshowItem *item;
    WkSlideshowItemType type;
    GSList *playlist = NULL;
    GRegex *regex;
    gchar **lines = NULL;
    gchar **p;
    int t;

    regex = g_regex_new ("^(.*) ([0-9]+)$", 0, 0, NULL);
    lines = g_strsplit (str, "\n", 0);
    if (g_strv_length (lines) == 0)
        goto out;

    for (p=lines+1; *p; p++)
    {
        char **tmp = NULL;

        tmp = g_regex_split (regex, *p, 0);
        if (g_strv_length (tmp) == 4)
        {
            t = atoi (tmp[2]);
            if (g_file_test (tmp[1]+7, G_FILE_TEST_IS_DIR))
            {
                append_dir_to_list (tmp[1]+7, t, &playlist);
                continue;
            }
            type = wk_slideshow_item_guess_type (tmp[1]);
            item = wk_slideshow_item_new (tmp[1], type, t, FALSE);
            playlist = g_slist_append (playlist, item);
        }
        g_strfreev (tmp);
    }

out:
    g_strfreev (lines);
    g_regex_unref (regex);

    return playlist;
}

static gboolean
wk_pl_loop_get_time (const gchar *start_time,
                     const gchar *end_time,
                     const gchar *weekdays,
                     WkPlLoopInfo *info)
{
    char *p;

    info->start_hour = -1;
    info->end_hour = -1;
    info->weekdays = 0;
    info->start_time[0] = 0;

    if (weekdays)
    {
        info->weekdays = 0;
        for (p=(char *)weekdays;*p;p++)
        {
            if (*p >= '0' && *p <= '6')
                info->weekdays |= 1 << (*p - '0');
        }
    }

    if (!start_time)
        return TRUE;

    if (2 != sscanf (start_time, "%d:%d", &info->start_hour, &info->start_min))
    {
        info->start_hour = -1;
        return FALSE;
    }

    /* if no end defined, or badly defined, end is 23:59 */
    if (2 != sscanf (end_time, "%d:%d", &info->end_hour, &info->end_min))
    {
        info->end_hour = 23;
        info->end_min = 59;
    }

    /* same if end is 00:00 */
    if ((info->end_hour == 0) && (info->end_min == 0))
    {
        info->end_hour = 23;
        info->end_min = 59;
    }

    /* if end after start? */
    if ((info->end_hour * 60 + info->end_min)
            <= (info->start_hour * 60 + info->start_min))
    {
        return FALSE;
    }

    /* make sure to have a correct format in the strings */
    snprintf (info->start_time, 6, "%02d:%02d",
              info->start_hour, info->start_min);
    snprintf (info->end_time, 6, "%02d:%02d",
              info->end_hour, info->end_min);

    return TRUE;
}

static gboolean
wk_pl_loop_get_date (const gchar *start_date,
                     const gchar *end_date,
                     WkPlLoopInfo *info)
{
    info->start_year = -1;
    info->end_year = -1;

    if (!start_date || !end_date)
        return TRUE;

    if (3 != sscanf (start_date, "%d/%d/%d",
                     &info->start_year, &info->start_month, &info->start_day))
    {
        info->start_year = -1;
        return FALSE;
    }

    if (3 != sscanf (end_date, "%d/%d/%d",
                     &info->end_year, &info->end_month, &info->end_day))
    {
        info->end_year = -1;
        return FALSE;
    }

    if (strcmp (start_date, end_date) > 0)
        return FALSE;

    /* make sure to have a correct format in the strings */
    snprintf (info->start_date, 11, "%04d/%02d/%02d",
              info->start_year, info->start_month, info->start_day);
    snprintf (info->end_date, 11, "%04d/%02d/%02d",
              info->end_year, info->end_month, info->end_day);

    return TRUE;
}

static gboolean
wk_pl_loop_get_from_markup (const gchar *start_time,
                            const gchar *end_time,
                            const gchar *weekdays,
                            const gchar *start_date,
                            const gchar *end_date,
                            WkPlLoopInfo *info)
{
    if (!wk_pl_loop_get_date (start_date, end_date, info)
        || !wk_pl_loop_get_time (start_time, end_time, weekdays, info))
    {
        return FALSE;
    }

    return TRUE;
}

void
wk_pl_info_print (WkPlLoopInfo *info)
{
    int i;

    if (info->start_year > 0)
    {
        printf ("Start: %s %s\n", info->start_date, info->start_time);
        if (info->end_year > 0)
            printf ("End  : %s %s\n", info->end_date, info->end_time);
    }
    else
    {
        if (info->start_hour >= 0)
        {
            printf ("Start: %s\n", info->start_time);
            printf ("End  : %s\n", info->end_time);
        }
    }

    if (info->weekdays > 0)
    {
        printf ("Days : %d (", info->weekdays);
        for (i=0;i<7;i++)
            if (info->weekdays & 1<<i)
                printf ("%d ", i);
        printf ("\b)\n");
    }
}

int
wk_pl_loop_detail_level (WkPlLoopInfo *info)
{
    int ret = 0;
    ret += (info->weekdays > 0)    ? 1 : 0;
    ret += (info->start_hour >= 0) ? 1 : 0;
    ret += (info->start_year >= 0) ? 1 : 0;

    DEBUG ("prio value: %d", ret);

    return ret;
}

gboolean
wk_pl_loop_fits_now (WkPlLoopInfo *info)
{
    time_t t_now;
    struct tm *now;
    gchar now_date[11];
    gchar now_time[6];
    gboolean ret = TRUE;

    t_now = time (NULL);
    now = localtime (&t_now);
    strftime (now_date, 11, "%Y/%m/%d", now);
    strftime (now_time, 6, "%H:%M", now);

    DEBUG ("Now: %s %s", now_date, now_time);

    /* check dates */
    if (info->start_year >= 0)
    {
        if ((strcmp (info->start_date, now_date) > 0)
            || (strcmp (now_date, info->end_date) > 0))
        {
            DEBUG ("Date not matching");
            ret = FALSE;
        }

    }

    /* check day of week */
    if (info->weekdays > 0 )
    {
        if ((info->weekdays & (1 << now->tm_wday)) == 0)
        {
            DEBUG ("Week day not matching");
            ret = FALSE;
        }
    }

    /* check the time */
    if (info->start_hour >= 0)
    {
        if ((strcmp (info->start_time, now_time) > 0)
            || (strcmp (now_time, info->end_time) > 0))
        {
            DEBUG ("%s %s %s", info->start_time, now_time, info->end_time);
            DEBUG ("Time not matching");
            ret = FALSE;
        }
    }

    return ret;
}

static void
start_element (GMarkupParseContext *context,
               const gchar         *element_name,
               const gchar        **attribute_names,
               const gchar        **attribute_values,
               gpointer             user_data,
               GError             **error)
{
    const gchar *start_date = NULL;;
    const gchar *end_date = NULL;
    const gchar *start_time = NULL;
    const gchar *end_time = NULL;
    const gchar *weekdays = NULL;
    const gchar *id = NULL;

    WkPlLoopInfo info;

    int i, prio;

    if (strcmp (element_name, "loop") != 0)
        return;

    for (i=0; attribute_names[i] != NULL; i++)
    {
        if ((0 == strcmp (attribute_names[i], "start"))
                && (attribute_values[i][0] != 0))
        {
            start_time = attribute_values[i];
        }
        else if ((0 == strcmp (attribute_names[i], "end"))
                && (attribute_values[i][0] != 0))
        {
            end_time = attribute_values[i];
        }
        else if ((0 == strcmp (attribute_names[i], "weekdays"))
                && (attribute_values[i][0] != 0))
        {
            weekdays = attribute_values[i];
        }
        else if  ((0 == strcmp (attribute_names[i], "from"))
                && (attribute_values[i][0] != 0))
        {
            start_date = attribute_values[i];
        }
        else if  ((0 == strcmp (attribute_names[i], "to"))
                && (attribute_values[i][0] != 0))
        {
            end_date = attribute_values[i];
        }
        else if (0 == strcmp (attribute_names[i], "id"))
        {
            id = attribute_values[i];
        }
    }

    DEBUG ("loop infos: id => %s, start_time => %s, stop_time => %s, "
           "weekdays => %s, from => %s, to => %s",
           id, start_time, end_time, weekdays, start_date, end_date);

    if (!id)
        return;

    DEBUG ("testing prio for id %s", id);
    if (! wk_pl_loop_get_from_markup (start_time, end_time, weekdays,
                                      start_date, end_date, &info))
    {
        DEBUG ("NOT valid");
        return;
    }

    DEBUG ("testing fit_now for id %s", id);
    if (! wk_pl_loop_fits_now (&info))
    {
        DEBUG ("NOT fitting");
        return;
    }

    prio = wk_pl_loop_detail_level (&info);
    if (prio > cur_priority)
    {
        cur_priority = prio;
        cur_id = atoi (id);
    }
}

static const GMarkupParser
first_run_parser = {
    start_element,
    NULL,
    NULL,
    NULL,
    NULL
};

static void
start_element_list (GMarkupParseContext *context,
                    const gchar         *element_name,
                    const gchar        **attribute_names,
                    const gchar        **attribute_values,
                    gpointer             user_data,
                    GError             **error)
{
    int i, id = -1;
    const gchar *uri = NULL;
    WkSlideshowItemType type = WK_SLIDESHOW_ITEM_TYPE_UNKNOWN;
    int t = 0;
    gboolean href = FALSE;
    GSList **l = (GSList **) user_data;
    GSList *tmp = NULL;

    if (! in_loop && ! idle)
    {
        if (strcmp (element_name, "loop") != 0)
            return;

        for (i=0; attribute_names[i] != NULL; i++)
        {
            if (0 == strcmp (attribute_names[i], "id"))
                id = atoi (attribute_values[i]);
            if (0 == strcmp (attribute_names[i], "idle")
                && (0 == strcmp (attribute_values[i], "1")))
            {
                idle = TRUE;
            }
        }

        if (id != cur_id)
        {
            idle = FALSE;
            return;
        }

        if (idle)
        {
            tmp = g_slist_append (*l, wk_slideshow_item_new (
                                        "idle", WK_SLIDESHOW_ITEM_TYPE_IDLE,
                                        60, FALSE));
            *l = tmp;
        }

        in_loop = TRUE;
        return;
    }

    for (i=0; attribute_names[i] != NULL; i++)
    {
        if (0 == strcmp (attribute_names[i], "uri"))
        {
            uri = attribute_values[i];
        }
        else if (0 == strcmp (attribute_names[i], "time"))
        {
            t = atoi (attribute_values[i]);
        }
        else if (0 == strcmp (attribute_names[i], "type"))
        {
            if (strcmp (attribute_values[i], "image") == 0)
                type = WK_SLIDESHOW_ITEM_TYPE_IMAGE;
            else if (strcmp (attribute_values[i], "video") == 0)
                type = WK_SLIDESHOW_ITEM_TYPE_VIDEO;
            else if (strcmp (attribute_values[i], "document") == 0)
                type = WK_SLIDESHOW_ITEM_TYPE_TEXT;
            else
                type = WK_SLIDESHOW_ITEM_TYPE_WEB;
        }
        else if (0 == strcmp (attribute_names[i], "href"))
        {
            href = (strcmp (attribute_values[i], "1") == 0) ? TRUE : FALSE;
        }
    }

    if (!uri)
        return;

    if (type == WK_SLIDESHOW_ITEM_TYPE_UNKNOWN)
        type = wk_slideshow_item_guess_type (uri);

    if (!t)
        t= DEFAULT_ITEM_TIMEOUT;

    if (uri && t)
    {
        gchar *absolute_uri;

        WkSettings *settings = GET_SETTINGS ();
        if (uri[0] == '/')
            absolute_uri = g_strconcat (settings->slideshow_file,
                                        "/resource", uri, NULL);
        else
            absolute_uri = g_strdup (uri);

        tmp = g_slist_append (*l, wk_slideshow_item_new (absolute_uri, type, t, href));
        *l = tmp;
        g_free (absolute_uri);
    }
}

static void
end_element_list (GMarkupParseContext *context,
                  const gchar         *element_name,
                  gpointer             user_data,
                  GError             **error)
{
    if (in_loop && (0 == strcmp (element_name, "loop")))
    {
        in_loop = FALSE;
    }
}

static const GMarkupParser
second_run_parser = {
    start_element_list,
    end_element_list,
    NULL,
    NULL,
    NULL
};

void
wk_pl_free_playlist_item (gpointer data, gpointer userdata)
{
    wk_slideshow_item_free ((WkSlideshowItem *) data);
}

void
wk_pl_free_playlist (GSList *playlist)
{
    g_slist_foreach (playlist, wk_pl_free_playlist_item, NULL);
    g_slist_free (playlist);
}

void
wk_pl_reset ()
{
    /* will force the download of data */
    prev_id = -1;
}

GSList *
wk_pl_get_playlist (const gchar *xml)
{
    GMarkupParseContext *context;
    GSList *playlist = NULL;

    cur_priority = 0;
    cur_id = 0;

    /* first parsing defines which loop id to use */
    context = g_markup_parse_context_new (&first_run_parser, 0, NULL, NULL);
    if (!g_markup_parse_context_parse (context, xml, strlen (xml), NULL))
    {
        g_markup_parse_context_free (context);

        /* Use the "old" syntax (uri timeout) */
        playlist = old_school_playlist_parser (xml);
        goto end;
    }

    DEBUG ("selected loop id: %d", cur_id);

    /* if no change in the playlist, do nothing */
    if (prev_id == cur_id)
        return NULL;

    prev_id = cur_id;
    idle = FALSE;

    /* second parsing gets the actual playlist */
    context = g_markup_parse_context_new (&second_run_parser, 0, &playlist, NULL);
    g_markup_parse_context_parse (context, xml, strlen (xml), NULL);

    DEBUG ("playlist len: %d", g_slist_length (playlist));

end:
    return playlist;
}

